import pandas as pd
import numpy as np
import math
import matplotlib.pyplot as plt

#The predictions per value on the validation dataset csv files
models = [
'validation_prediction_results_long.csv'
,
'validation_prediction_results_short_v3.csv'
,
'validation_prediction_results_non_TV2_short_v3.csv'
]



#Iterates over the prediction values for all three models, calculates metrics the given value and appends results to dictionary, which is then plotted
model_metrics = {}

for model in models:
    val_data = pd.read_csv(model)
        
    val_data.dropna(inplace=True)

    values = []
    TP = {}
    FP = {}
    TN = {}
    FN = {}
    total = []

    for i in range(0, val_data.shape[0]):
        frame = val_data.iloc[i]

        if frame['Label'] == 'nan' or frame['Label'] == 'Nan':
            continue

        if frame['Prediction'] > 0:
                
            if str(frame['Label']) == str(1.0):
                total.append(0)


    for value in np.arange(0,1.01,0.01):

        values.append(value)
        TPs = []
        FPs = []
        TNs = []
        FNs = []

        for i in range(0, val_data.shape[0]):
            frame = val_data.iloc[i]

            if frame['Label'] == 'Nan' or frame['Label'] == 'nan':
                continue

            if frame['Prediction'] > value:
                
                if str(frame['Label']) == str(1.0):
                    TPs.append(frame['Text'])
                else:
                    FPs.append(frame['Text'])
            else:
                if str(frame['Label']) == str(1.0):
                    FNs.append(frame['Text'])
                else:
                    TNs.append(frame['Text'])

                    
        TP[str(value)] = TPs
        FP[str(value)] = FPs
        TN[str(value)] = TNs
        FN[str(value)] = FNs


    total = len(total)  

    TPes = []
    TNes = []
    FPes = []
    FNes = []
    accs = []
    errs = []
    precisions = []
    recalls = []
    hmfhs = []
    hmrs = []
    dist = []
    tests = []
    F1s = []
    TPRs = []
    FPRs = []
    base = []

    for ind, val in enumerate(values):

        TPv = len(TP[str(val)])
        TNv = len(TN[str(val)])
        FPv = len(FP[str(val)])
        FNv = len(FN[str(val)])

        
        try:
            TPes.append(1-(1/TPv))
        except:
            TPes.append(0.0)

        try:
            FPes.append(1-(1/FPv))
        except:
            FPes.append(0.0)

        try:
            FNes.append(1-(1/FNv))
        except:
            FNes.append(0.0)

        try:
            d = 1-(1/(abs(FPv-TPv)))
            dist.append(d)
        except:
            dist.append(d)

        try:
            hmfh = TPv/FPv
            hmfhs.append(hmfh)
        except:
            hmfhs.append(0.0)

        try:
            test = TPv/FPv
            tests.append(test)
        except:
            tests.append(0.0)
            

        try:
            hmr = (TPv/FNv)/10
            hmrs.append(hmr)
        except:
            hmrs.append(0.0)      

        try:
            acc = (TPv+TNv)/(TPv+TNv+FPv+FNv)
            accs.append(acc)
        except:
            accs.append(0.0)   

        try:
            precision = TPv / (TPv + FPv)
            precisions.append(precision)
        except:
            precisions.append(0.0)

        try:
            recall = TPv / (TPv + FNv)
            recalls.append(recall)
        except:
            recalls.append(0.0)
            
        try:
            F1 = 2*precision*recall / (precision + recall)
            F1s.append(F1)
        except:
            F1s.append(0.0)


        try:
            TPR = TPv/(TPv+FNv)
            TPRs.append(TPR)
        except:
            TPRs.append(0.0)

        try:
            FPR = FPv/(FPv+TNv)
            FPRs.append(TPR)
        except:
            FPRs.append(0.0)

        try:
            base.append(1-val)
        except:
            pass
            


    model_metrics[model] = {
    'TP':TPes,
    'TN':TNes,
    'FP':FPes,
    'FN':FNes,
    'acc':accs,
    'err':errs,
    'precision':precisions,
    'recall':recalls,
    'F1':F1s,
    'dist':dist,
    'TPR':TPRs,
    'FPR':FPRs,
    'base':base
    }
    print(models.index(model)+1,'/','3')


#Plot the models' results
model_names = [
'TV2FacebookCoronaPoliticsIronyLongNorBERT2',
'TV2FacebookCoronaPoliticsIronyNorBERT2',
'TV2FacebookIronyNorBERT2'
]

for ind, model in enumerate(models):
    #print(model_metrics[model])

    #plt.plot(values, model_metrics[model]['TP'], label='True Positives'+' '+model_names[ind])
    #plt.plot(values, model_metrics[model]['FP'], label='False Positives'+' '+model_names[ind])
    #plt.plot(values, model_metrics[model]['dist'], label='TP and FP Distance'+' '+model_names[ind])
    #plt.plot(values, model_metrics[model]['F1'], label='F1 Score'+' '+model_names[ind])
    #plt.plot(values, model_metrics[model]['precision'], label='Precision'+' '+model_names[ind])
    #plt.plot(values, model_metrics[model]['recall'], label='Recall'+' '+model_names[ind])
    #plt.plot(values, model_metrics[model]['acc'], label='Accuracy'+' '+model_names[ind])
    #plt.plot(values, model_metrics[model]['TPR'], label='TP Rate'+' '+model_names[ind])
    #plt.plot(values, model_metrics[model]['TPR'], label='FP Rate'+' '+model_names[ind])
    #plt.plot(model_metrics[model]['FPR'], values, label='TP Rate'+' '+model_names[ind])
    #plt.plot(model_metrics[model]['base'], values)
    #plt.plot(model_metrics[model]['TPR'],model_metrics[model]['FPR'], label='ROC'+' '+model_names[ind])
    


#plt.ylabel('True Positive Rate')
#plt.xlabel('False Positive Rate')
plt.axis([0, 1.0, 0, 1.0])
#plt.axis([0, 0.25, 0, 1.0])

#plt.legend(loc='upper right')
plt.legend(loc='lower right')
#plt.legend(loc='upper left')

plt.xlabel('Percentage of Metric')
plt.ylabel('Values')
plt.show()





    
