#Code modified from https://curiousily.com/posts/multi-label-text-classification-with-bert-and-pytorch-lightning/

def main():
    import pandas as pd
    import numpy as np
    from tqdm.auto import tqdm
    import torch
    import torch.nn as nn
    from torch.utils.data import Dataset, DataLoader
    from transformers import BertTokenizerFast, AutoTokenizer, BertModel, get_linear_schedule_with_warmup
    import pytorch_lightning as pl
    from torch.optim import AdamW
    from pytorch_lightning import LightningModule
    from torchmetrics.functional import accuracy, f1, auroc
    from pytorch_lightning.callbacks import ModelCheckpoint, EarlyStopping
    from pytorch_lightning.loggers import TensorBoardLogger
    from sklearn.model_selection import train_test_split
    from sklearn.metrics import classification_report, multilabel_confusion_matrix
    from sklearn import metrics
    import seaborn as sns
    from pylab import rcParams
    import os
    from transformers.models.longformer.modeling_longformer import LongformerSelfAttention
    from transformers import LongformerTokenizerFast, LongformerForSequenceClassification, AutoTokenizer, LongformerTokenizer, BertForSequenceClassification, AutoModelForSequenceClassification , AutoModel, BertForSequenceClassification
    #from fairscale.nn import checkpoint_wrapper
    import fairscale
    #from pytorch_lightning.plugins import DDPPlugin, DDPShardedPlugin
    from fairscale.nn import checkpoint_wrapper, auto_wrap, wrap
    from fairscale.nn import checkpoint_wrapper
    from pytorch_lightning.strategies import DDPStrategy, DDPShardedStrategy
    #from pytorch_lightning.strategies.
    from pytorch_lightning.strategies import DeepSpeedStrategy
    import dill as pickle
    ################################################################


    path = 'D:/noBert/Context/Results/'
    max_token_length = 1024
    batch_size = 1
    model_name = 'LONGnorBERT_1024_v2'
    n_epochs = 30
    worker_amount = 0

    LABEL_COLUMNS = ['label']

    #os.environ["CUDA_VISIBLE_DEVICES"]="1"
    os.environ["no_deprecation_warning"]="True"
    os.environ["PL_TORCH_DISTRIBUTED_BACKEND"]="gloo"

    #device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    tokenizer = BertTokenizerFast.from_pretrained('D:/noBert/LongNorBERT_v2')
     
    #Create Random Dataset
    ###################################################################################################
    train_data_path = 'D:\\noBert\\Context_Dataset_Single_Label.csv'
    train = pd.read_csv(train_data_path)

    train['label'] = train['ironic']


    seperate_article = False

    if not seperate_article:
        #Combine Article, Comment and Reply text into one field with special seperation tokens
        for i in range(0,train.shape[0]):
            frame = train.iloc[i]
            
            article_text = frame['article']
            comment_text = frame['comment']
            reply_text = frame['reply']
            
            if not str(reply_text) == 'nan':
                text = article_text.rstrip() +' [SEP] '+ comment_text.rstrip() +' [SEP] '+reply_text.rstrip()
            else:
                text = article_text.rstrip() +' [SEP] '+ comment_text.rstrip()

            train.at[i,'text'] = text
    else:
        #Combine Comment and Reply text into one field with special seperation tokens
        for i in range(0,train.shape[0]):
            frame = train.iloc[i]
            
            comment_text = frame['comment']
            reply_text = frame['reply']
            
            if not str(reply_text) == 'nan':
                text = comment_text.rstrip() +' [SEP] '+reply_text.rstrip()
            else:
                text = comment_text.rstrip()

            train.at[i,'text'] = text


    if not seperate_article:
        train = train[['text', 'label']].reset_index(drop=True)
    else:
        train = train[['article', 'text', 'label']].reset_index(drop=True)
        

    if False:
        #Investigate data
        for i in range(0,train.shape[0]):
            frame = train.iloc[i]
            print(frame['text'], frame['label'])
        input()



    #Remove too long
    for i in range(0,train.shape[0]):
        frame = train.iloc[i]

        if len(tokenizer(frame['text'])[0]) > max_token_length:
            train.at[i,'text'] = 'nan'
            train.at[i,'label'] = 'nan'

    train.dropna(inplace=True)

    

    print(str(train['label'].value_counts()[1.0])+'/'+str(len(train['label'])))

    print(str((train['label'].value_counts()[1.0]/len(train['label']))*100),'Percent are ironic')



    train_size = 0.8
    train_data=train.sample(frac=train_size)
    test_dataset=train.drop(train_data.index).reset_index(drop=True)
    train_dataset = train_data.reset_index(drop=True)



    print('Train Dataset Irony Count: '+str(train_dataset['label'].value_counts()[1.0]))
    print('Test Dataset Irony Count: '+str(test_dataset['label'].value_counts()[1.0]))



    train_dataset.to_csv('train_single_long_data.csv', encoding='utf-8-sig', index=False)
    test_dataset.to_csv('test_single_long_data.csv', encoding='utf-8-sig', index=False)


    train_data = pd.read_csv('train_single_long_data.csv')
    test_data = pd.read_csv('test_single_long_data.csv')

    train_data.dropna(inplace=True)
    test_data.dropna(inplace=True)
    
    print(train_data)
    print(test_data)




    class norBERTLongSelfAttention(LongformerSelfAttention):
        def forward(
            self,
            hidden_states,
            attention_mask=None,
            head_mask=None,
            encoder_hidden_states=None,
            encoder_attention_mask=None,
            past_key_value=None,
            output_attentions=False,
        ):
            """
            :class:`LongformerSelfAttention` expects `len(hidden_states)` to be multiple of `attention_window`. Padding to
            `attention_window` happens in :meth:`LongformerModel.forward` to avoid redoing the padding on each layer.

            Modified LongformerSelfAttention layer to work with RoBERTa models. They have different set of arguments
            compared to original Longformer attention: RoBERTa is passing some additional arguments and arguments required
            by LongformerSelfAttention are missing so we have to add it manually.

            The `attention_mask` is changed in :meth:`LongformerModel.forward` from 0, 1, 2 to:

                * -10000: no attention
                * 0: local attention
                * +10000: global attention
            """
            hidden_states = hidden_states.transpose(0, 1)

            # project hidden states
            query_vectors = self.query(hidden_states)
            key_vectors = self.key(hidden_states)
            value_vectors = self.value(hidden_states)

            ## Lines below has been changed

            attention_mask = attention_mask.squeeze(dim=2).squeeze(dim=1)

            # is index masked or global attention
            is_index_masked = attention_mask < 0
            is_index_global_attn = attention_mask > 0
            is_global_attn = any(is_index_global_attn.flatten())

            ## End of modification

            seq_len, batch_size, embed_dim = hidden_states.size()
            assert (
                embed_dim == self.embed_dim
            ), f"hidden_states should have embed_dim = {self.embed_dim}, but has {embed_dim}"

            # normalize query
            query_vectors /= math.sqrt(self.head_dim)

            query_vectors = query_vectors.view(seq_len, batch_size, self.num_heads, self.head_dim).transpose(0, 1)
            key_vectors = key_vectors.view(seq_len, batch_size, self.num_heads, self.head_dim).transpose(0, 1)

            attn_scores = self._sliding_chunks_query_key_matmul(
                query_vectors, key_vectors, self.one_sided_attn_window_size
            )

            # values to pad for attention probs
            
            ## Lines below has been changed
            
            remove_from_windowed_attention_mask = (attention_mask != 0).unsqueeze(dim=-1).unsqueeze(dim=-1)

            ## End of modification

            # cast to fp32/fp16 then replace 1's with -inf
            float_mask = remove_from_windowed_attention_mask.type_as(query_vectors).masked_fill(
                remove_from_windowed_attention_mask, -10000.0
            )
            # diagonal mask with zeros everywhere and -inf inplace of padding
            diagonal_mask = self._sliding_chunks_query_key_matmul(
                float_mask.new_ones(size=float_mask.size()), float_mask, self.one_sided_attn_window_size
            )

            # pad local attention probs
            attn_scores += diagonal_mask

            assert list(attn_scores.size()) == [
                batch_size,
                seq_len,
                self.num_heads,
                self.one_sided_attn_window_size * 2 + 1,
            ], f"local_attn_probs should be of size ({batch_size}, {seq_len}, {self.num_heads}, {self.one_sided_attn_window_size * 2 + 1}), but is of size {attn_scores.size()}"

            # compute local attention probs from global attention keys and contact over window dim
            if is_global_attn:
                # compute global attn indices required through out forward fn
                (
                    max_num_global_attn_indices,
                    is_index_global_attn_nonzero,
                    is_local_index_global_attn_nonzero,
                    is_local_index_no_global_attn_nonzero,
                ) = self._get_global_attn_indices(is_index_global_attn)
                # calculate global attn probs from global key

                global_key_attn_scores = self._concat_with_global_key_attn_probs(
                    query_vectors=query_vectors,
                    key_vectors=key_vectors,
                    max_num_global_attn_indices=max_num_global_attn_indices,
                    is_index_global_attn_nonzero=is_index_global_attn_nonzero,
                    is_local_index_global_attn_nonzero=is_local_index_global_attn_nonzero,
                    is_local_index_no_global_attn_nonzero=is_local_index_no_global_attn_nonzero,
                )
                # concat to local_attn_probs
                # (batch_size, seq_len, num_heads, extra attention count + 2*window+1)
                attn_scores = torch.cat((global_key_attn_scores, attn_scores), dim=-1)

                # free memory
                del global_key_attn_scores

            attn_probs = F.softmax(attn_scores, dim=-1, dtype=torch.float32)  # use fp32 for numerical stability

            # softmax sometimes inserts NaN if all positions are masked, replace them with 0
            attn_probs = torch.masked_fill(attn_probs, is_index_masked[:, :, None, None], 0.0)
            attn_probs = attn_probs.type_as(attn_scores)

            # free memory
            del attn_scores

            # apply dropout
            attn_probs = F.dropout(attn_probs, p=self.dropout, training=self.training)

            value_vectors = value_vectors.view(seq_len, batch_size, self.num_heads, self.head_dim).transpose(0, 1)

            # compute local attention output with global attention value and add
            if is_global_attn:
                # compute sum of global and local attn
                attn_output = self._compute_attn_output_with_global_indices(
                    value_vectors=value_vectors,
                    attn_probs=attn_probs,
                    max_num_global_attn_indices=max_num_global_attn_indices,
                    is_index_global_attn_nonzero=is_index_global_attn_nonzero,
                    is_local_index_global_attn_nonzero=is_local_index_global_attn_nonzero,
                )
            else:
                # compute local attn only
                attn_output = self._sliding_chunks_matmul_attn_probs_value(
                    attn_probs, value_vectors, self.one_sided_attn_window_size
                )

            assert attn_output.size() == (batch_size, seq_len, self.num_heads, self.head_dim), "Unexpected size"
            attn_output = attn_output.transpose(0, 1).reshape(seq_len, batch_size, embed_dim).contiguous()

            # compute value for global attention and overwrite to attention output
            # TODO: remove the redundant computation
            if is_global_attn:
                global_attn_output, global_attn_probs = self._compute_global_attn_output_from_hidden(
                    hidden_states=hidden_states,
                    max_num_global_attn_indices=max_num_global_attn_indices,
                    is_local_index_global_attn_nonzero=is_local_index_global_attn_nonzero,
                    is_index_global_attn_nonzero=is_index_global_attn_nonzero,
                    is_local_index_no_global_attn_nonzero=is_local_index_no_global_attn_nonzero,
                    is_index_masked=is_index_masked,
                )

                # get only non zero global attn output
                nonzero_global_attn_output = global_attn_output[
                    is_local_index_global_attn_nonzero[0], :, is_local_index_global_attn_nonzero[1]
                ]

                # overwrite values with global attention
                attn_output[is_index_global_attn_nonzero[::-1]] = nonzero_global_attn_output.view(
                    len(is_local_index_global_attn_nonzero[0]), -1
                )
                # The attention weights for tokens with global attention are
                # just filler values, they were never used to compute the output.
                # Fill with 0 now, the correct values are in 'global_attn_probs'.
                attn_probs[is_index_global_attn_nonzero] = 0

            outputs = (attn_output.transpose(0, 1),)

            if output_attentions:
                outputs += (attn_probs,)

            return outputs + (global_attn_probs,) if (is_global_attn and output_attentions) else outputs



    class norBERTLongForSequenceClassification(BertForSequenceClassification):
        def __init__(self, config):
            super().__init__(config)
            #print(nn.ModuleList([norBERTLongSelfAttention(config, layer_id=index)]))
            #self.bert.encoder.layer = nn.ModuleList([norBERTLongSelfAttention(config, layer_id=index) for index in range(config.num_hidden_layers)])
            for index in range(config.num_hidden_layers):
                print(index)
                norBERTLongSelfAttention(config, layer_id=index)
                

    class norBERTLongModel(BertModel):
        def __init__(self, config):
            super().__init__(config)
            self.encoder.layer = nn.ModuleList([norBERTLongSelfAttention(config, layer_id=index) for index in range(config.num_hidden_layers)])

            

    class IronyCommentsDataset(Dataset):

      def __init__(
        self, 
        data: pd.DataFrame, 
        tokenizer: BertTokenizerFast, 
        max_token_len: int = max_token_length
      ):
        self.tokenizer = tokenizer
        self.data = data
        self.max_token_len = max_token_len
        
      def __len__(self):
        return len(self.data)

      def __getitem__(self, index: int):
        data_row = self.data.iloc[index]

        text = data_row.text
        label = data_row[LABEL_COLUMNS]

        encoding = self.tokenizer(
          text,
          add_special_tokens=True,
          max_length=self.max_token_len,
          return_token_type_ids=False,
          padding="max_length",
          truncation=True,
          return_attention_mask=True,
          return_tensors='pt',
        )

        return dict(
          #text=text,
          input_ids=encoding["input_ids"].flatten(),
          attention_mask=encoding["attention_mask"].flatten(),
          label=torch.FloatTensor(label)
        )



    class IronyLongCommentDataModule(pl.LightningDataModule):

      def __init__(self, train_df, test_df, tokenizer, batch_size=1, max_token_len=max_token_length):
        super().__init__()
        self.batch_size = 1
        self.train_df = train_df
        self.test_df = test_df
        self.tokenizer = tokenizer
        self.max_token_len = max_token_len

      def setup(self, stage=None):
        self.train_dataset = IronyCommentsDataset(
          self.train_df,
          self.tokenizer,
          self.max_token_len
        )

        self.test_dataset = IronyCommentsDataset(
          self.test_df,
          self.tokenizer,
          self.max_token_len
        )

      def train_dataloader(self):
        return DataLoader(
          self.train_dataset,
          batch_size=self.batch_size,
          shuffle=True,
          #pin_memory=True,
          num_workers=worker_amount
        )

      def val_dataloader(self):
        return DataLoader(
          self.test_dataset,
          batch_size=self.batch_size,
          #pin_memory=True,
          num_workers=worker_amount
        )

      def test_dataloader(self):
        return DataLoader(
          self.test_dataset,
          batch_size=self.batch_size,
          #pin_memory=True,
          num_workers=worker_amount
        )



    data_module = IronyLongCommentDataModule(
      train_data,
      test_data,
      tokenizer,
      batch_size=1,
      max_token_len=max_token_length
    )




    class IronyCommentTagger(pl.LightningModule):

      def __init__(self, n_classes: int, n_training_steps=None, n_warmup_steps=None):
        super().__init__()

        self.bert = norBERTLongForSequenceClassification.from_pretrained('D:/noBert/LongNorBERT_v2', attention_window=512, num_labels=1, return_dict=True)
        self.classifier = nn.Linear(self.bert.config.hidden_size, n_classes)
        self.n_training_steps = n_training_steps
        self.n_warmup_steps = n_warmup_steps
        self.criterion = nn.BCELoss()

      def forward(self, input_ids, attention_mask, label=None):

        output = self.bert(input_ids, attention_mask=attention_mask)
        output = torch.sigmoid(output[0])
        loss = 0
        if label is not None:
            loss = self.criterion(output, label)

        return loss, output

      def training_step(self, batch, batch_idx):
        input_ids = batch["input_ids"]
        attention_mask = batch["attention_mask"]
        label = batch["label"]
        loss, outputs = self(input_ids, attention_mask, label)
        self.log("train_loss", loss, prog_bar=True, logger=True)
        return {"loss": loss, "predictions": outputs.detach(), "label": label}

      def validation_step(self, batch, batch_idx):
        input_ids = batch["input_ids"]
        attention_mask = batch["attention_mask"]
        label = batch["label"]
        loss, outputs = self(input_ids, attention_mask, label)
        self.log("val_loss", loss, prog_bar=True, logger=True)
        return loss

      def test_step(self, batch, batch_idx):
        input_ids = batch["input_ids"]
        attention_mask = batch["attention_mask"]
        label = batch["label"]
        loss, outputs = self(input_ids, attention_mask, label)
        self.log("test_loss", loss, prog_bar=True, logger=True)
        return loss

      def training_epoch_end(self, outputs):
        label = []
        predictions = []
        for output in outputs:
          for out_label in output["label"].detach().cpu():
            label.append(out_label)
          for out_predictions in output["predictions"].detach().cpu():
            predictions.append(out_predictions)

        label = torch.stack(label).int()
        predictions = torch.stack(predictions)

        for i, name in enumerate(LABEL_COLUMNS):
          class_roc_auc = auroc(predictions[:, i], label[:, i])
          self.logger.experiment.add_scalar(f"{name}_roc_auc/Train", class_roc_auc, self.current_epoch)


      def configure_optimizers(self):
        optimizer = AdamW(self.parameters(), lr=2e-5)
        scheduler = get_linear_schedule_with_warmup(
          optimizer,
          num_warmup_steps=self.n_warmup_steps,
          num_training_steps=self.n_training_steps
        )
        return dict(
            optimizer=optimizer,
            lr_scheduler=dict(scheduler=scheduler,interval='step'), 
        )


    dummy_model = nn.Linear(2, 1)

    optimizer = AdamW(params=dummy_model.parameters(), lr=0.001)

    warmup_steps = 20
    total_training_steps = 100

    scheduler = get_linear_schedule_with_warmup(
      optimizer, 
      num_warmup_steps=warmup_steps,
      num_training_steps=total_training_steps
    )

    learning_rate_history = []

    for step in range(total_training_steps):
      optimizer.step()
      scheduler.step()
      learning_rate_history.append(optimizer.param_groups[0]['lr'])


    steps_per_epoch=len(train_data) // 1
    total_training_steps = steps_per_epoch * n_epochs

    warmup_steps = total_training_steps // 5

    model = IronyCommentTagger(
      n_classes=len(LABEL_COLUMNS),
      n_warmup_steps=warmup_steps,
      n_training_steps=total_training_steps 
    )


    checkpoint_callback = ModelCheckpoint(
      dirpath="D:/noBert/Checkpoints",
      filename="best-checkpoint",
      save_top_k=3,
      verbose=True,
      monitor="val_loss",
      mode="min"
    )

    logger = TensorBoardLogger("lightning_logs", name="irony")

    trainer = pl.Trainer(
        strategy="ddp_sharded",
        gpus=2,
        accumulate_grad_batches=1,
        #precision=16,
        accelerator='gpu',
        logger=logger,
        checkpoint_callback=checkpoint_callback,
        #callbacks=[early_stopping_callback],
        max_epochs=n_epochs,
        #num_sanity_val_steps=0,
        progress_bar_refresh_rate=1
    )

    

    trainer.fit(model, data_module)


if __name__ == "__main__":
    main()

#Potential improvements
#Enable precision 16
#Change from automodel to longformer in custom class, and change self.bert to whatever should be instead
