#Code modified from https://curiousily.com/posts/multi-label-text-classification-with-bert-and-pytorch-lightning/

def main():
    import pandas as pd
    import numpy as np
    from tqdm.auto import tqdm
    import torch
    import torch.nn as nn
    from torch.utils.data import Dataset, DataLoader
    from transformers import BertTokenizerFast, AutoTokenizer, BertModel, get_linear_schedule_with_warmup
    import pytorch_lightning as pl
    from torch.optim import AdamW
    from torchmetrics.functional import accuracy, f1, auroc
    from pytorch_lightning.callbacks import ModelCheckpoint, EarlyStopping
    from pytorch_lightning.loggers import TensorBoardLogger
    from sklearn.model_selection import train_test_split
    from sklearn.metrics import classification_report, multilabel_confusion_matrix
    from sklearn import metrics
    import seaborn as sns
    from pylab import rcParams
    import os
    import fairscale
    #from pytorch_lightning.plugins import DDPPlugin, DDPShardedPlugin
    from fairscale.nn import checkpoint_wrapper, auto_wrap, wrap
    from fairscale.nn import checkpoint_wrapper
    from pytorch_lightning.strategies import DDPStrategy, DDPShardedStrategy
    #from pytorch_lightning.strategies.
    from pytorch_lightning.strategies import DeepSpeedStrategy
    import dill as pickle

    ################################################################


    path = 'D:/noBert/Context/Results/'
    max_token_length = 512
    batch_size = 1
    model_name = 'ShortNorBERT_512_v1'
    n_epochs = 10

    if not os.path.exists(path+model_name):
        os.makedirs(path+model_name)
        os.makedirs(path+model_name+'/Saved_Model_Folder')

    os.environ["CUDA_VISIBLE_DEVICES"]="1"
    os.environ["no_deprecation_warning"]="True"
    #os.environ["PL_TORCH_DISTRIBUTED_BACKEND"]="gloo"
    
    #device = torch.device("cuda:1" if torch.cuda.is_available() else "cpu")

    tokenizer = BertTokenizerFast.from_pretrained('D:/noBert/norBERT2')

    #Create Random Dataset
    ###################################################################################################
    train_data_path = 'D:\\noBert\\Context_Dataset_Single_Label.csv'
    train = pd.read_csv(train_data_path)

    train['label'] = train['ironic']

    duplicate = []

    #Seperate rows
    for i in range(0,train.shape[0]):
        frame = train.iloc[i]
        
        comment_text = frame['comment']
        reply_text = frame['reply']
        
        if not str(reply_text) == 'nan':
            text = reply_text#.rstrip()
        else:
            text = comment_text#.rstrip()

        encoding = tokenizer(text, return_token_type_ids=False)
  

        if len(encoding['input_ids']) > max_token_length:
            print('Removed')
            continue

        
        if text in duplicate:
            continue
        duplicate.append(text)
        train.at[i,'text'] = text
        

    train = train[['text', 'label']].reset_index(drop=True)




    print(str(train['label'].value_counts()[1.0])+'/'+str(len(train['label'])))

    print(str((train['label'].value_counts()[1.0]/len(train['label']))*100),'Percent are ironic')

    train.dropna(inplace=True)

    train_size = 0.8
    train_data=train.sample(frac=train_size)
    test_dataset=train.drop(train_data.index).reset_index(drop=True)
    train_dataset = train_data.reset_index(drop=True)



    print('Train Dataset Irony Count: '+str(train_dataset['label'].value_counts()[1.0]))
    print('Test Dataset Irony Count: '+str(test_dataset['label'].value_counts()[1.0]))

    
    LABEL_COLUMNS = ['label']

    input()
    train_dataset.to_csv('train_single_short_data.csv', encoding='utf-8-sig', index=False)
    test_dataset.to_csv('test_single_short_data.csv', encoding='utf-8-sig', index=False)

    if False:
        #Investigate data
        for i in range(0,test_dataset.shape[0]):
            frame = test_dataset.iloc[i]
            print(i, frame['text'], frame['label'])


    train_data = pd.read_csv('train_single_short_data.csv')
    test_data = pd.read_csv('test_single_short_data.csv')
    
    print(train_data)
    print(test_data)

    
    tokenizer = BertTokenizerFast.from_pretrained('D:/noBert/norBERT2')

    #print(tokenizer)

    #input()

    class IronyCommentsDataset(Dataset):

      def __init__(
        self, 
        data: pd.DataFrame, 
        tokenizer: BertTokenizerFast, 
        max_token_len: int = max_token_length
      ):
        self.tokenizer = tokenizer
        self.data = data
        self.max_token_len = max_token_len
        
      def __len__(self):
        return len(self.data)

      def __getitem__(self, index: int):
        data_row = self.data.iloc[index]

        text = data_row.text
        label = data_row[LABEL_COLUMNS]

        encoding = self.tokenizer(
          text,
          add_special_tokens=True,
          max_length=self.max_token_len,
          return_token_type_ids=False,
          padding="max_length",
          truncation=True,
          return_attention_mask=True,
          return_tensors='pt',
        )

        return dict(
          #text=text,
          input_ids=encoding["input_ids"].flatten(),
          attention_mask=encoding["attention_mask"].flatten(),
          label=torch.FloatTensor(label)
        )



    class IronyCommentDataModule(pl.LightningDataModule):

      def __init__(self, train_df, test_df, tokenizer, batch_size=1, max_token_len=512):
        super().__init__()
        self.batch_size = 1
        self.train_df = train_df
        self.test_df = test_df
        self.tokenizer = tokenizer
        self.max_token_len = max_token_len

      def setup(self, stage=None):
        self.train_dataset = IronyCommentsDataset(
          self.train_df,
          self.tokenizer,
          self.max_token_len
        )

        self.test_dataset = IronyCommentsDataset(
          self.test_df,
          self.tokenizer,
          self.max_token_len
        )

      def train_dataloader(self):
        return DataLoader(
          self.train_dataset,
          batch_size=self.batch_size,
          shuffle=True,
          num_workers=0
        )

      def val_dataloader(self):
        return DataLoader(
          self.test_dataset,
          batch_size=self.batch_size,
          num_workers=0
        )

      def test_dataloader(self):
        return DataLoader(
          self.test_dataset,
          batch_size=self.batch_size,
          num_workers=0
        )



    data_module = IronyCommentDataModule(
      train_data,
      test_data,
      tokenizer,
      batch_size=1,
      max_token_len=max_token_length
    )



    class IronyCommentTagger(pl.LightningModule):

      def __init__(self, n_classes: int, n_training_steps=None, n_warmup_steps=None):
        super().__init__()
        self.bert = BertModel.from_pretrained('D:/noBert/norBERT2', return_dict=True)
        self.classifier = nn.Linear(self.bert.config.hidden_size, n_classes)
        self.n_training_steps = n_training_steps
        self.n_warmup_steps = n_warmup_steps
        self.criterion = nn.BCELoss()

      def forward(self, input_ids, attention_mask, label=None):
        output = self.bert(input_ids, attention_mask=attention_mask)
        output = self.classifier(output.pooler_output)
        output = torch.sigmoid(output)    
        loss = 0
        if label is not None:
            loss = self.criterion(output, label)
        return loss, output

      def training_step(self, batch, batch_idx):
        input_ids = batch["input_ids"]
        attention_mask = batch["attention_mask"]
        label = batch["label"]
        loss, outputs = self(input_ids, attention_mask, label)
        self.log("train_loss", loss, prog_bar=True, logger=True)
        return {"loss": loss, "predictions": outputs.detach(), "label": label}

      def validation_step(self, batch, batch_idx):
        input_ids = batch["input_ids"]
        attention_mask = batch["attention_mask"]
        label = batch["label"]
        loss, outputs = self(input_ids, attention_mask, label)
        self.log("val_loss", loss, prog_bar=True, logger=True)
        return loss

      def test_step(self, batch, batch_idx):
        input_ids = batch["input_ids"]
        attention_mask = batch["attention_mask"]
        label = batch["label"]
        loss, outputs = self(input_ids, attention_mask, label)
        self.log("test_loss", loss, prog_bar=True, logger=True)
        return loss

      def training_epoch_end(self, outputs):
        
        label = []
        predictions = []
        for output in outputs:
          for out_label in output["label"].detach().cpu():
            label.append(out_label)
          for out_predictions in output["predictions"].detach().cpu():
            predictions.append(out_predictions)

        label = torch.stack(label).int()
        predictions = torch.stack(predictions)

        for i, name in enumerate(LABEL_COLUMNS):
          class_roc_auc = auroc(predictions[:, i], label[:, i])
          self.logger.experiment.add_scalar(f"{name}_roc_auc/Train", class_roc_auc, self.current_epoch)


      def configure_optimizers(self):

        optimizer = AdamW(self.parameters(), lr=2e-5)

        scheduler = get_linear_schedule_with_warmup(
          optimizer,
          num_warmup_steps=self.n_warmup_steps,
          num_training_steps=self.n_training_steps
        )

        return dict(
          optimizer=optimizer,
          lr_scheduler=dict(
            scheduler=scheduler,
            interval='step'
          )
        )


    dummy_model = nn.Linear(2, 1)

    optimizer = AdamW(params=dummy_model.parameters(), lr=0.001)

    warmup_steps = 20
    total_training_steps = 100

    scheduler = get_linear_schedule_with_warmup(
      optimizer, 
      num_warmup_steps=warmup_steps,
      num_training_steps=total_training_steps
    )

    learning_rate_history = []

    for step in range(total_training_steps):
      optimizer.step()
      scheduler.step()
      learning_rate_history.append(optimizer.param_groups[0]['lr'])


    steps_per_epoch=len(train_data) // 1
    total_training_steps = steps_per_epoch * n_epochs

    warmup_steps = total_training_steps // 5

    print(total_training_steps)
    print(warmup_steps)

    model = IronyCommentTagger(
      n_classes=len(LABEL_COLUMNS),
      n_warmup_steps=warmup_steps,
      n_training_steps=total_training_steps 
    )


    checkpoint_callback = ModelCheckpoint(
      dirpath="D:/noBert/Checkpoints",
      filename="best-checkpoint",
      save_top_k=3,
      verbose=True,
      monitor="val_loss",
      mode="min"
    )

    logger = TensorBoardLogger("lightning_logs", name="Irony-comments")

    trainer = pl.Trainer(
      #strategy="ddp_sharded",
      gpus=1,
      logger=logger,
      #accumulate_grad_batches=1,
      accelerator='gpu',
      checkpoint_callback=checkpoint_callback,
      #callbacks=[early_stopping_callback],
      max_epochs=n_epochs,
      progress_bar_refresh_rate=1
    )

    trainer.fit(model, data_module)


if __name__ == "__main__":
    main()
